/*
 * SCUT - A command line tool to manipulate shortcuts
 *
 * Copyright (c) 2004-2016 Malcolm Smith
 * malxau@users.sourceforge.net
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

//
//  For once, intentionally not WIN32_LEAN_AND_MEAN
//

#include <windows.h>
#include <shlobj.h>
#include <objidl.h>
#ifndef MINICRT
#include <stdio.h>
#include <stdlib.h>
#endif
#include <tchar.h>

#ifdef MINICRT
#include <minicrt.h>
#endif

#include "msupd.h"

typedef enum _operations {
    OPERATION_UNKNOWN=0,
    OPERATION_CREATE,
    OPERATION_MODIFY,
    OPERATION_EXEC,
    OPERATION_DUMP
} operations;

const
CHAR strLicenseText[] = 
     "\n"
     "Copyright (c) 2004-16 Malcolm J. Smith\n"
     "\n"
     "Permission is hereby granted, free of charge, to any person obtaining a copy\n"
     "of this software and associated documentation files (the \"Software\"), to deal\n"
     "in the Software without restriction, including without limitation the rights\n"
     "to use, copy, modify, merge, publish, distribute, sublicense, and/or sell\n"
     "copies of the Software, and to permit persons to whom the Software is\n"
     "furnished to do so, subject to the following conditions:\n"
     "\n"
     "The above copyright notice and this permission notice shall be included in\n"
     "all copies or substantial portions of the Software.\n"
     "\n"
     "THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\n"
     "IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\n"
     "FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\n"
     "AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\n"
     "LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\n"
     "OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN\n"
     "THE SOFTWARE.\n";


void usage(TCHAR * app)
{
    _tprintf(_T("%s -?\n"),app);
    _tprintf(_T("%s -create filename [-target target] [-args args] [-desc description]\n")
             _T("  [-hotkey hotkey] [-iconpath filename [-iconindex num]]\n")
             _T("  [-show showcmd] [-workingdir workingdir]\n"),
             app);
    _tprintf(_T("%s -modify filename [-target target] [-args args] [-desc description]\n")
             _T("  [-hotkey hotkey] [-iconpath filename [-iconindex num]]\n")
             _T("  [-show showcmd] [-workingdir workingdir]\n"),
             app);
    _tprintf(_T("%s -exec filename [-target target] [-args args] [-show showcmd]\n")
             _T("  [-workingdir workingdir]\n"),
             app);
    _tprintf(_T("%s -dump filename\n"), app);

    _tprintf(_T("\n"));

    _tprintf(_T("%s -get-update    (Update to latest version of current type)\n"), app);
    _tprintf(_T("%s -get-daily     (Get the latest daily version)\n"), app);
    _tprintf(_T("%s -get-source    (Get source for this version)\n"), app);
    _tprintf(_T("%s -get-stable    (Get the latest stable version)\n"), app);
    _tprintf(_T("%s -get-symbol    (Get debugging symbol for this version)\n"), app);
    _tprintf(_T("%s -get-ver-x.y   (Get a specific version of the program)\n"), app);

    _tprintf(_T("%hs"), strLicenseText);

    exit(EXIT_SUCCESS);
}

typedef enum {
    ScutUpdateTypeNone = 0,
    ScutUpdateTypeBinary,
    ScutUpdateTypeSource,
    ScutUpdateTypeSymbol,
    ScutUpdateTypeStable,
    ScutUpdateTypeDaily
} SCUT_UPDATE_TYPE;

BOOL
ScutAssembleUrl(
    LPTSTR szUrl,
    DWORD szUrlLength,
    LPTSTR szVerString,
    LPTSTR szFileName,
    LPTSTR szFileExtension
    )
{
    TCHAR szTempFileName[128];

    if (szFileName == NULL) {
        _stprintf_s(szTempFileName,
                    sizeof(szTempFileName)/sizeof(szTempFileName[0]),
                    _T("%s%s/%s"),
#ifdef _M_AMD64
                    _T("amd64"),
#elif defined(_M_IX86)
                    _T("win32"),
#endif
#ifdef UNICODE
                    _T("-unicode"),
#else
                    _T("-ansi"),
#endif
                    _T("scut")
                    );

        szFileName = szTempFileName;
    }

    _stprintf_s(szUrl,
                szUrlLength,
                _T("http://www.malsmith.net/download/?obj=%s/%s/%s%s"),
                _T("scut"),
                szVerString?szVerString:
#if SCUT_BUILD_ID
                _T("latest-daily"),
#else
                _T("latest-stable"),
#endif
                szFileName,
                szFileExtension?szFileExtension:_T(".exe")
                );

    return TRUE;
}


BOOL
ScutUpdateBinary(
    SCUT_UPDATE_TYPE UpdateType,
    LPTSTR VersionToUpdateTo
    )
{
    TCHAR szAgent[128];
    TCHAR szUrl[250];
    TCHAR szVer[32];
    LPTSTR TargetName = NULL;
    UpdError Error = UpdErrorSuccess;

#if SCUT_BUILD_ID
    _stprintf_s(szVer, sizeof(szVer)/sizeof(szVer[0]), _T("%i.%i.%i"), SCUT_VER_MAJOR, SCUT_VER_MINOR, SCUT_BUILD_ID);
#else
    _stprintf_s(szVer, sizeof(szVer)/sizeof(szVer[0]), _T("%i.%i"), SCUT_VER_MAJOR, SCUT_VER_MINOR);
#endif

    _stprintf_s(szAgent, sizeof(szAgent)/sizeof(szAgent[0]), _T("scut %s"), szVer);

    if (UpdateType == ScutUpdateTypeBinary) {
        ScutAssembleUrl(szUrl, sizeof(szUrl)/sizeof(szUrl[0]), VersionToUpdateTo, NULL, NULL);
    } else if (UpdateType == ScutUpdateTypeSource) {
        ScutAssembleUrl(szUrl, sizeof(szUrl)/sizeof(szUrl[0]), szVer, _T("scut-source"), _T(".zip"));
        TargetName = _T("scut-source.zip");
    } else if (UpdateType == ScutUpdateTypeSymbol) {
        ScutAssembleUrl(szUrl, sizeof(szUrl)/sizeof(szUrl[0]), szVer, NULL, _T(".pdb"));
        TargetName = _T("scut.pdb");
    } else if (UpdateType == ScutUpdateTypeStable) {
        ScutAssembleUrl(szUrl, sizeof(szUrl)/sizeof(szUrl[0]), _T("latest-stable"), NULL, NULL);
    } else if (UpdateType == ScutUpdateTypeDaily) {
        ScutAssembleUrl(szUrl, sizeof(szUrl)/sizeof(szUrl[0]), _T("latest-daily"), NULL, NULL);
    }

    _tprintf(_T("Updating from %s\n"), szUrl);
    Error = UpdateBinaryFromUrl(szUrl, TargetName, szAgent);
    _tprintf(_T("Update result: %s\n"), UpdateErrorString(Error));
    return TRUE;
}

int _tmain(int argc, TCHAR * argv[])
{
    operations op         = OPERATION_UNKNOWN;
    TCHAR * szFile        = NULL;

    TCHAR * szArgs        = NULL;
    TCHAR * szDesc        = NULL;
    WORD    wHotkey       = (WORD)-1;
    TCHAR * szIcon        = NULL;
    WORD    wIcon;
    WORD    wShow         = (WORD)-1;
    TCHAR * szTarget      = NULL;    
    TCHAR * szWorkingDir  = NULL;
    TCHAR * app           = argv[0];
    IShellLink *scut      = NULL;
#ifndef UNICODE
    wchar_t szUniTarget[MAX_PATH];
#endif
    HRESULT hRes;
    BOOL bHandled;
    SCUT_UPDATE_TYPE UpdateType = ScutUpdateTypeNone;
    LPTSTR VersionToUpdateTo = NULL;


    if (argc < 2 ||
        _tcscmp(argv[1], _T("/?")) == 0 ||
        _tcscmp(argv[1], _T("-?")) == 0 ||
        _tcscmp(argv[1], _T("--help")) == 0) {

        usage(app);
    }

    argc--;
    argv++;
    
    while (argc) {

        bHandled = FALSE;

        if (_tcscmp(argv[0], _T("-get-daily")) == 0) {
            bHandled = TRUE;
            UpdateType = ScutUpdateTypeDaily;
        }

        if (_tcscmp(argv[0], _T("-get-source")) == 0) {
            bHandled = TRUE;
            UpdateType = ScutUpdateTypeSource;
        }

        if (_tcscmp(argv[0], _T("-get-stable")) == 0) {
            bHandled = TRUE;
            UpdateType = ScutUpdateTypeStable;
        }

        if (_tcscmp(argv[0], _T("-get-symbol")) == 0) {
            bHandled = TRUE;
            UpdateType = ScutUpdateTypeSymbol;
        }

        if (_tcscmp(argv[0], _T("-get-update")) == 0) {
            bHandled = TRUE;
            UpdateType = ScutUpdateTypeBinary;
        }

        if (_tcsncmp(argv[0], _T("-get-ver-"), 9) == 0) {
            bHandled = TRUE;
            UpdateType = ScutUpdateTypeBinary;
            VersionToUpdateTo = &argv[0][9];
        }

        if (bHandled) {
            argc--;
            argv++;
            continue;
        }

        //
        // Every other parameter needs an argument
        //

        if (argc == 1) {
            usage(app);
        }

        if (_tcscmp(argv[0], _T("-create")) == 0) {
            op = OPERATION_CREATE;
            szFile = argv[1];
            bHandled = TRUE;
        }
        if (_tcscmp(argv[0], _T("-modify")) == 0) {
            op = OPERATION_MODIFY;
            szFile = argv[1];
            bHandled = TRUE;
        }
        if (_tcscmp(argv[0], _T("-exec")) == 0) {
            op = OPERATION_EXEC;
            szFile = argv[1];
            bHandled = TRUE;
        }
        if (_tcscmp(argv[0], _T("-dump")) == 0) {
            op = OPERATION_DUMP;
            szFile = argv[1];
            bHandled = TRUE;
        }


        if (_tcscmp(argv[0], _T("-args")) == 0) {
            szArgs = argv[1];
            bHandled = TRUE;
        }
        if (_tcscmp(argv[0], _T("-desc")) == 0) {
            szDesc = argv[1];
            bHandled = TRUE;
        }
        if (_tcscmp(argv[0], _T("-hotkey")) == 0) {
            wHotkey = _ttoi(argv[1]);
            bHandled = TRUE;
        }
        if (_tcscmp(argv[0], _T("-iconpath")) == 0) {
            szIcon = argv[1];
            bHandled = TRUE;
        }
        if (_tcscmp(argv[0], _T("-iconindex")) == 0) {
            wIcon = _ttoi(argv[1]);
            bHandled = TRUE;
        }

        if (_tcscmp(argv[0], _T("-show")) == 0) {
            wShow = _ttoi(argv[1]);
            bHandled = TRUE;
        }
        if (_tcscmp(argv[0], _T("-target")) == 0) {
            szTarget = argv[1];
            bHandled = TRUE;
        }
        if (_tcscmp(argv[0], _T("-workingdir")) == 0) {
            szWorkingDir = argv[1];
            bHandled = TRUE;
        }

        if (!bHandled) {
            usage(app);
        }

        argv+=2;
        argc-=2;
    }

    if (UpdateType != ScutUpdateTypeNone) {
        ScutUpdateBinary(UpdateType, VersionToUpdateTo);
        return EXIT_SUCCESS;
    }

    if (op == OPERATION_UNKNOWN) {
        usage(app);
    }

    hRes = CoInitialize(NULL);
    if (!SUCCEEDED(hRes)) {
        _ftprintf( stderr, _T("CoInitialize failure: %x\n"), hRes);
        return EXIT_FAILURE;
    }

    IPersistFile *savedfile;
#ifndef UNICODE
    MultiByteToWideChar(CP_ACP, 0, szFile, -1, szUniTarget, MAX_PATH);
#endif

    hRes = CoCreateInstance(CLSID_ShellLink,
        NULL,
        CLSCTX_INPROC_SERVER,
        IID_IShellLink,
        (void **)&scut);

    if (!SUCCEEDED(hRes)) {
        _ftprintf( stderr,
            _T("CoCreateInstance failure: %x\n"),
            hRes);
        return EXIT_FAILURE;
    }

    hRes = scut->QueryInterface(IID_IPersistFile, (void **)&savedfile);
    if (!SUCCEEDED(hRes)) {
        _ftprintf( stderr,
            _T("QueryInstance IPersistFile failure: %x\n"),
            hRes);
        return EXIT_FAILURE;
    }

    if (op == OPERATION_MODIFY ||
        op == OPERATION_EXEC ||
        op == OPERATION_DUMP) {
#ifndef UNICODE
        hRes = savedfile->Load(szUniTarget, TRUE);
#else
        hRes = savedfile->Load(szFile, TRUE);
#endif
        if (!SUCCEEDED(hRes)) {
            _ftprintf( stderr, _T("Load failure: %x\n"), hRes);
            return EXIT_FAILURE;
        }
    }

    if (op == OPERATION_DUMP) {
        TCHAR szTemp[MAX_PATH];
        int   wTemp;
        if (scut->GetPath(szTemp, MAX_PATH, NULL, 0) != NOERROR) {
            _ftprintf( stderr, _T("GetPath failure\n"));
            return EXIT_FAILURE;
        }
        _tprintf(_T("Target:          %s\n"), szTemp);
        if (scut->GetArguments(szTemp, MAX_PATH) != NOERROR) {
            _ftprintf( stderr, _T("GetArguments failure\n"));
            return EXIT_FAILURE;
        }
        _tprintf(_T("Arguments:       %s\n"), szTemp);
        if (scut->GetWorkingDirectory(szTemp, MAX_PATH) != NOERROR) {
            _ftprintf( stderr, _T("GetWorkingDirectory failure\n"));
            return EXIT_FAILURE;
        }
        _tprintf(_T("Working Dir:     %s\n"), szTemp);
        if (scut->GetDescription(szTemp, MAX_PATH) != NOERROR) {
            _ftprintf( stderr, _T("GetDescription failure\n"));
            return EXIT_FAILURE;
        }
        _tprintf(_T("Description:     %s\n"), szTemp);
        if (scut->GetIconLocation(szTemp, MAX_PATH, &wTemp) != NOERROR) {
            _ftprintf( stderr, _T("GetIconLocation failure\n"));
            return EXIT_FAILURE;
        }
        _tprintf(_T("Icon Path:       %s\n"), szTemp);
        _tprintf(_T("Icon Index:      %i\n"), wTemp);
        if (scut->GetShowCmd(&wTemp) != NOERROR) {
            _ftprintf( stderr, _T("GetShowCmd failure\n"));
            return EXIT_FAILURE;
        }
        _tprintf(_T("Show state:      %i\n"), wTemp);
        if (scut->GetHotkey(&wShow) != NOERROR) {
            _ftprintf( stderr, _T("GetHotkey failure\n"));
            return EXIT_FAILURE;
        }
        _tprintf(_T("Hotkey:          %i\n"), wShow);
        return EXIT_SUCCESS;
    }

    if (szTarget) {
        if (scut->SetPath(szTarget) != NOERROR) {
            _ftprintf( stderr, _T("SetPath failure\n"));
            return EXIT_FAILURE;
        }
    }

    if (szArgs) {
        if (scut->SetArguments(szArgs) != NOERROR) {
            _ftprintf( stderr, _T("SetArguments failure\n"));
            return EXIT_FAILURE;
        }
    }

    if (szDesc) {
        if (scut->SetDescription(szDesc) != NOERROR) {
            _ftprintf( stderr, _T("SetDescription failure\n"));
            return EXIT_FAILURE;
        }
    }

    if (wHotkey != (WORD)-1) {
        if (scut->SetHotkey(wHotkey) != NOERROR) {
            _ftprintf( stderr, _T("SetHotkey failure\n"));
            return EXIT_FAILURE;
        }
    }

    if (szIcon) {
        if (scut->SetIconLocation(szIcon, wIcon) != NOERROR) {
            _ftprintf( stderr, _T("SetIconLocation failure\n"));
            return EXIT_FAILURE;
        }
    }

    if (wShow != (WORD)-1) {
        if (scut->SetShowCmd(wShow) != NOERROR) {
            _ftprintf( stderr, _T("SetShowCmd failure\n"));
            return EXIT_FAILURE;
        }
    }

    if (szWorkingDir) {
        if (scut->SetWorkingDirectory(szWorkingDir) != NOERROR) {
            _ftprintf( stderr, _T("SetWorkingDirectory failure\n"));
            return EXIT_FAILURE;
        }
    }

    if (op == OPERATION_MODIFY ||
        op == OPERATION_CREATE) {
#ifndef UNICODE
        hRes = savedfile->Save(szUniTarget, TRUE);
#else
        hRes = savedfile->Save(szFile, TRUE);
#endif
        if (!SUCCEEDED(hRes)) {
            _ftprintf( stderr, _T("Save failure: %x\n"), hRes);
            return EXIT_FAILURE;
        }
    } else if (op == OPERATION_EXEC) {
        TCHAR szFile[MAX_PATH];
        TCHAR szArgs[MAX_PATH];
        TCHAR szDir[MAX_PATH];
        INT nShow;
        HINSTANCE hApp;
        if (scut->GetWorkingDirectory(szDir, MAX_PATH) != NOERROR) {
            _ftprintf( stderr, _T("GetWorkingDirectory failure\n"));
            return EXIT_FAILURE;
        }
        if (scut->GetArguments(szArgs,MAX_PATH) != NOERROR) {
            _ftprintf( stderr, _T("GetArguments failure\n"));
            return EXIT_FAILURE;
        }
        if (scut->GetPath(szFile, MAX_PATH, NULL, 0) != NOERROR) {
            _ftprintf( stderr, _T("GetPath failure\n"));
            return EXIT_FAILURE;
        }
        if (scut->GetShowCmd(&nShow) != NOERROR) {
            _ftprintf( stderr, _T("GetShowCmd failure\n"));
            return EXIT_FAILURE;
        }

        hApp = ShellExecute(NULL,
            NULL,
            szFile,
            szArgs,
            szDir,
            nShow);
        if ((UINT)hApp <= 32) {
            _ftprintf( stderr,
                _T("ShellExecute failure: %i\n"),
                hApp);
            return EXIT_FAILURE;
        }
    }

    return EXIT_SUCCESS;
}

// vim:sw=4:ts=4:et:
